/* **********************************************************
 *   Program:    PC Keyboard
 *   wb - 09.Mar.2005 -- ShiftLock-Behandlung 
 *   wb - 29.Jan.2005 -- Based on ...
 *        TOPDO  - Status-LEDs auf Tastatur
 *
 *   Created:    07.09.99 21:05
 *   Author:     V. Brajer - vlado.brajer@kks.s-net.net
 *               http://www.sparovcek.net/bray
 *               based on Atmel's AVR313 application note
 *   Comments:   AVRGCC port - original for IAR C
 *********************************************************** */
#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/signal.h>
#include <avr/wdt.h>
#include <avr/pgmspace.h>

#include "pckb.h"
#include "scancodes.h"

#define REAL_SHIFT_LOCK   /* Else shift-lock act like an additional shift key */

extern volatile char mode;   // mcu.c
/* -------------------- Portability macros ------------------------ */
// Keybord Input Data-Pin
#define DDR_KB  DDRD
#define PIN_KB  PIND
#define PORT_KB PORTD
#define KB_DATABITNR PD4
#define KB_DATABIT (1<<KB_DATABITNR)

// Keyboard Clocking
#define KB_CLOCKINT     INT1
#define KB_INTERRUPT    SIG_INTERRUPT1
#define KB_INT_PORT     PORTD
#define KB_INT_PINNO    PD3
#define FALLING_EDGE()  MCUCR = (MCUCR & ~((1<<ISC11)|(1<<ISC10)))|(1<<ISC11)|(0<<ISC10)
#define RISING_EDGE()   MCUCR |= (1<<ISC11)|(1<<ISC10)
#define ANY_EDGE()      MCUCR = (MCUCR & ~((1<<ISC11)|(1<<ISC10)))|(0<<ISC11)|(1<<ISC10)
#define KB_INT_PIN      (1<<KB_INT_PINNO)

/* -------------------- Data definitions ----------------------- */
#define BUFF_SIZE 16

#ifdef DEBUG_KB
unsigned char edge, bitcount;                // 0 = neg.  1 = pos.
unsigned int kb_buffer[BUFF_SIZE];
unsigned int *inpt, *outpt;
volatile unsigned char buffcnt;
//volatile unsigned int anyInt = 0;
#else
static unsigned char edge, bitcount;        // 0 = neg.  1 = pos.
static unsigned int kb_buffer[BUFF_SIZE];
static unsigned int *inpt, *outpt;
static volatile unsigned char buffcnt;
#endif

/* --------------------- Code ---------------------- */
void kb_init(void)
{
    GICR    &= ~(1<<KB_CLOCKINT);   // Disable INT1 Interrupt // GIMSK
    edge     = 0;                   // 0 = falling edge  1 = rising edge
    bitcount = 11;

    inpt =  kb_buffer;              // Initialize buffer
    outpt = kb_buffer;
    buffcnt = 0;
    DDR_KB  &= ~KB_DATABIT;         // Data-Pin as input
    PORT_KB |= KB_DATABIT;          // Pull-Up on Databit
    KB_INT_PORT |= KB_INT_PIN;      // Pull-Up on Interrupt Line (clocking)
    ANY_EDGE();                     // INT1 interrupt on any edge
    
    GIFR   = (1<<KB_CLOCKINT);      // Clear any pending Interrupt
    GICR  |= (1<<KB_CLOCKINT);      // Enable INT1 Interrupt // GIMSK
}

/* signal handler for external interrupt int1 */
SIGNAL(KB_INTERRUPT)
{
    static unsigned char data;                // Holds the received scan code
    static unsigned int  shift = 0;
    static struct 
    {
        unsigned is_up      : 1;
        unsigned edge       : 1;
        unsigned extend     : 1;     // E0
        unsigned shift_lock : 1;
        unsigned alt        : 1;
        unsigned alt_gr     : 1;
        unsigned left_ctrl  : 1;
        unsigned right_ctrl : 1;
    } flags = { 0,0,0,0,0,0,0,0 };

    //anyInt++;
    if (!flags.edge)                         // Routine entered at falling edge
    {
        if(bitcount < 11 && bitcount > 2)    // Bit 3 to 10 is data. Parity bit,
        {                                    // start and stop bits are ignored.
            data >>= 1;
            if((PIN_KB & KB_DATABIT) != 0)
                data |= 0x80;
        }
        flags.edge = 1;
    }
    else
    {                                        // Routine entered at rising edge
        flags.edge = 0;
        if(--bitcount == 0)                  // All bits received
        {
            bitcount = 11;
            // decode somewhat
            if (!flags.is_up)               // Last data received was the up-key identifier
            {
                //outp(sc,PORTB);           // show me scancode - for debuging
                switch (data)
                {
                case 0xF0:                  // The up-key identifier
                    flags.is_up = 1;
                    break;
                case 0xE0:                  // Extended Character
                    shift |= EXTENDED;
                    flags.extend = 1;
                    break;
                case 0x58:                  // SHIFT-LOCK
#ifdef REAL_SHIFT_LOCK
                    shift ^= LOCK_SHIFT;
                    flags.shift_lock ^= 1;
                    // now, toggle the shift lock LED on the keyboard
#else                
                    shift |= LOCK_SHIFT;
                    flags.shift_lock = 1;
#endif                    
                    break;
                case 0x12:                  // Left SHIFT
                    shift |= LEFT_SHIFT;
                    // now, toggle the shift lock LED on the keyboard
                    shift &= ~LOCK_SHIFT;
                    flags.shift_lock = 0;
                    break;
                case 0x59:                  // Right SHIFT
                    shift |= RIGHT_SHIFT;
                    // now, toggle the shift lock LED on the keyboard
                    shift &= ~LOCK_SHIFT;
                    flags.shift_lock = 0;
                    break;

                case 0x14:                  // Ctrl
                    if(flags.extend)
                    {
                        shift |= RIGHT_CTRL;
                        flags.right_ctrl = 1;
                    }
                    else
                    {
                        shift |= LEFT_CTRL;
                        flags.left_ctrl = 1;
                    }
                    break;
                case 0x11:                  // Alt
                    if(flags.extend)
                    {
                        shift |= RIGHT_ALT;
                        flags.alt_gr = 1;
                    }
                    else
                    {
                        shift |= LEFT_ALT;
                        flags.alt = 1;
                    }
                    break;
#if 0
                case 0x??:                  // Windows-Key
                    shift |= WINDOWS_KEY;
                    break;
#endif
                default:
                    if(flags.alt && flags.left_ctrl)    // ctrl-alt+Nummber = display-Mode-switch
                    { 
                        switch(data)
                        {
                        case 0x45: mode = 0; break;   // <strg><alt>0
                        case 0x16: mode = 1; break;   // <strg><alt>1
                        case 0x1E: mode = 2; break;   // ...
                        case 0x26: mode = 3; break;
                        case 0x25: mode = 4; break;
                        case 0x2E: mode = 5; break;
                        case 0x36: mode = 6; break;
                        case 0x3D: mode = 7; break;
                        case 0x3E: mode = 8; break;
                        case 0x46: mode = 9; break;
                        default:
                            goto putInBuffer;
                        }                   
                    } 
                    else
                    {
                    putInBuffer:

                        if (buffcnt < BUFF_SIZE)                    // If buffer not full
                        {
                            *inpt = (shift | data);                 // Put scancode into buffer
                            inpt++;                                 // Increment pointer
                            buffcnt++;
                            if (inpt >= (kb_buffer + BUFF_SIZE))
                                inpt = kb_buffer;
                        }                    
//                      else Beep();
                    }                    
                    shift &= ~(EXTENDED);
                    flags.extend = 0;
                    break;
                }
            }
            else        //    
            {
                flags.is_up = 0;        // Two 0xF0 in a row not allowed
                switch (data)
                {
                case 0xE0:              // Extended Character
                    shift &= ~EXTENDED;
                    flags.extend = 0;
                    break;
                case 0xF0:              // The up-key identifier
                    flags.is_up = 1;
                    break;
#ifdef REAL_SHIFT_LOCK
                case 0x58:              // SHIFT-LOCK
                    break;   // ignore
#else
                case 0x58:              // SHIFT-LOCK
                    shift &= ~(LOCK_SHIFT);
                    flags.shift_lock = 0;
                    break;
#endif                    
                case 0x12:              // Left SHIFT
                    shift &= ~LEFT_SHIFT;
                    break;
                case 0x59:              // Right SHIFT
                    shift &= ~RIGHT_SHIFT;
                    break;

                case 0x14:              // Ctrl
                    if(flags.extend)
                    {
                        shift &= ~RIGHT_CTRL;
                        flags.right_ctrl = 0;
                        flags.extend = 0;
                    }
                    else
                    {
                        shift &= ~LEFT_CTRL;
                        flags.left_ctrl = 0;
                    }
                    break;
                case 0x11:              // Alt
                    if(flags.extend)
                    {
                        shift &= ~RIGHT_ALT;
                        flags.alt_gr = 0;
                        flags.extend = 0;
                    }
                    else
                    {
                        shift &= ~LEFT_ALT;
                        flags.alt = 0;
                    }
                    break;
#if 0
                case 0x??:              // Windows-Key
                    shift &= ~WINDOWS_KEY;
                    break;
#endif
//--                case 0x76 : lcd_cls(); break;         // ESC - clear my LCD
                }
            }
        }
    }
}


unsigned int kb_decode(unsigned int data)
{
    unsigned char sc, c;
    prog_char *z;

    z = (prog_char *)&scanCodes[0];
    sc = data & 0xFF;
    data &= 0xFF00;
    while((c = pgm_read_byte(z)) != 0)
    {
        if(c == sc)
            break;
        z += 4;
    }
    if(sc == c)
    {   
        z++;
        if (data & RIGHT_ALT)                               // Alt Gr
        {
            // [3]
            return (data | pgm_read_byte(&z[2]));
        }
        if (data & (LEFT_CTRL|RIGHT_CTRL))                  // Ctrl
        {
            // [1]
            sc = pgm_read_byte(z);
            if(sc >= 'a' && sc <= 'z')
                return (data | (sc - 'a' + 1));                    
        }
        if (data & (LEFT_SHIFT|RIGHT_SHIFT|LOCK_SHIFT))     // Shifted
            z++;
        return (data | pgm_read_byte(z));                   // or Unshifted
    }
    return data;
}

unsigned int kb_peekcode(void)
{
    if(buffcnt == 0)             
        return 0;
    return (*outpt);                              // Get scancode    
}

unsigned int kb_peekchar(void)
{
    if(buffcnt == 0)             
        return 0;
    return (kb_decode(*outpt));                  // Get character    
}


unsigned int kb_getcode(void)
{
    unsigned int word;

    while(buffcnt == 0)                         // Wait for data
        wdt_reset();
    word = *outpt;                              // Get scancode
    outpt++;                                    // Increment pointer
    if (outpt >= (kb_buffer + BUFF_SIZE))
        outpt = kb_buffer;
    buffcnt--;                                  // Decrement buffer count
    return word;
}


unsigned int kb_getchar(void)
{
    return kb_decode(kb_getcode());
}

unsigned char kbhit(void)
{
    return buffcnt;
}

void kb_flush(void)
{
    unsigned char sreg = SREG;
    cli();
    inpt    = kb_buffer;                      // Re-Initialize buffer
    outpt   = kb_buffer;
    buffcnt = 0;
    SREG = sreg;
}

